/*
Copyright (c) 2013-2014, Shodruky Rhyammer
Copyright (c) 2014, Trygve Aaberge
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

	Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

	Redistributions in binary form must reproduce the above copyright notice, this
	list of conditions and the following disclaimer in the documentation and/or
	other materials provided with the distribution.

	Neither the name of the copyright holders nor the names of its
	contributors may be used to endorse or promote products derived from
	this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Originally code for Mandelbrot computation, licenced as stated above, from:
https://github.com/parallella/parallella-examples/tree/master/mandelbrot
Used as basis and modified for this Game of Life implementation.
*/


#include "e_lib.h"
#include "shared_data.h"

#define BUF_ADDRESS 0x8f000000
#define PAGE_OFFSET 0x2000
#define PAGE_SIZE 0x2000

unsigned int row, col, core;
unsigned int xres, yres;

unsigned int fetch_node(char *src, int x, int y, int x_diff, int y_diff) {
	int get_global = 0;
	int col_new = col;
	int row_new = row;
	int x_new = x + x_diff;
	if (x_new < 0) {
		col_new--;
		if (col_new < 0) {
			return 0;
		}
		x_new = xres - 1;
		get_global = 1;
	} else if (x_new == xres) {
		col_new++;
		if (col_new >= COLS) {
			return 0;
		}
		x_new = 0;
		get_global = 1;
	}
	int y_new = y + y_diff;
	if (y_new < 0) {
		row_new--;
		if (row_new < 0) {
			return 0;
		}
		y_new = yres - 1;
		get_global = 1;
	} else if (y_new == yres) {
		row_new++;
		if (row_new >= ROWS) {
			return 0;
		}
		y_new = 0;
		get_global = 1;
	}

	char *src2 = src + (y_new * yres + x_new) * 4;
	if (get_global) {
		src2 = e_get_global_address(row_new, col_new, src2);
	}
	return *src2;
}

int main(void)
{
	e_coreid_t coreid;
	coreid = e_get_coreid();
	e_coords_from_coreid(coreid, &row, &col);
	core = row * COLS + col;
	unsigned int frame = 1;
	unsigned int page = 1;
	volatile msg_block_t *msg = (msg_block_t *)BUF_ADDRESS;
	xres = msg->info.xres;
	yres = msg->info.yres;

	// For simplicity, each value is stored in a word. Since a word is 4 bytes,
	// and the only possible values are 0 and 1, it would be possible to store 32
	// values for each word.

	char *src[2] = {(char *)(PAGE_OFFSET), (char *)(PAGE_OFFSET + PAGE_SIZE)};
	unsigned int *pixel_last = (unsigned int *)src[0];
	unsigned int *pixel_cur = (unsigned int *)src[1];
	unsigned int x, y;
	for (y = 0; y < yres; y++) {
		for (x = 0; x < xres; x++) {
			*pixel_last = msg->pixels[core][y][x];
			*pixel_cur = 0;
			pixel_last++;
			pixel_cur++;
		}
	}

	while (1) {
		msg->msg_d2h[core].coreid = coreid;
		msg->msg_d2h[core].value = frame;
		__asm__ __volatile__ ("trap 4");

		pixel_cur = (unsigned int *)src[page];
		page = page ^ 1;

		if (frame > FRAMES) {
			break;
		}
		frame++;

		for (y = 0; y < yres; y++) {
			for (x = 0; x < xres; x++) {
				int is_live = fetch_node(src[page], x, y, 0, 0);
				int nr_live = 0;
				nr_live += fetch_node(src[page], x, y, -1, -1);
				nr_live += fetch_node(src[page], x, y, -1, 0);
				nr_live += fetch_node(src[page], x, y, -1, 1);
				nr_live += fetch_node(src[page], x, y, 0, -1);
				nr_live += fetch_node(src[page], x, y, 0, 1);
				nr_live += fetch_node(src[page], x, y, 1, -1);
				nr_live += fetch_node(src[page], x, y, 1, 0);
				nr_live += fetch_node(src[page], x, y, 1, 1);
				if (is_live) {
					if (nr_live < 2 || nr_live > 3) {
						*pixel_cur = 0;
					} else {
						*pixel_cur = 1;
					}
				} else if (nr_live == 3) {
					*pixel_cur = 1;
				} else {
					*pixel_cur = is_live;
				}

				pixel_cur++;
			}
		}
	}

	pixel_cur = (unsigned int *)src[page];
	for (y = 0; y < yres; y++) {
		for (x = 0; x < xres; x++) {
			msg->pixels[core][y][x] = *pixel_cur;
			pixel_cur++;
		}
	}

	msg->msg_d2h[core].finished = 1;

	return 0;
}
