/*
Copyright (c) 2013-2014, Shodruky Rhyammer
Copyright (c) 2014, Trygve Aaberge
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

	Redistributions of source code must retain the above copyright notice, this
	list of conditions and the following disclaimer.

	Redistributions in binary form must reproduce the above copyright notice, this
	list of conditions and the following disclaimer in the documentation and/or
	other materials provided with the distribution.

	Neither the name of the copyright holders nor the names of its
	contributors may be used to endorse or promote products derived from
	this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Originally code for Mandelbrot computation, licenced as stated above, from:
https://github.com/parallella/parallella-examples/tree/master/mandelbrot
Used as basis and modified for this Game of Life implementation.
*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <stdint.h>
#include <time.h>
#include <e-hal.h>
#include "shared_data.h"

#define BUF_OFFSET 0x01000000

static inline void nano_wait(uint32_t sec, uint32_t nsec)
{
	struct timespec ts;
	ts.tv_sec = sec;
	ts.tv_nsec = nsec;
	nanosleep(&ts, NULL);
}

int main(int argc, char *argv[])
{
	e_platform_t eplat;
	e_epiphany_t edev;
	e_mem_t emem;
	static msg_block_t msg;
	memset(&msg, 0, sizeof(msg));
	struct timespec time;
	double time0, time1;

	e_init(NULL);
	e_reset_system();
	e_get_platform_info(&eplat);
	e_alloc(&emem, BUF_OFFSET, sizeof(msg_block_t));

	unsigned int xres = XRES;
	unsigned int yres = YRES;
	msg.info.xres = xres;
	msg.info.yres = yres;

	unsigned int core = 0;
	unsigned int row = 0;
	unsigned int col = 0;
	volatile unsigned int vepiphany[CORES];
	volatile unsigned int vcoreid = 0;
	unsigned int vhost[CORES];
	for (core = 0; core < CORES; core++) {
		vepiphany[core] = 0;
		vhost[core] = 0;
	}

	FILE *file = fopen("grid.txt", "r");
	if (file == NULL) {
		fprintf(stderr, "Could not open grid.txt\n");
		exit(1);
	}

	int x, y, c;
	int eol = 0;
	for (row = 0; row < ROWS; row++) {
		for (y = 0; y < yres; y++) {
			for (col = 0; col < COLS; col++) {
				core = row * COLS + col;
				for (x = 0; x < xres; x++) {
					if (!eol) {
						c = fgetc(file);
						if (c == '\n') {
							eol = 1;
						}
					}
					if (c == '1') {
						msg.pixels[core][y][x] = 1;
					} else {
						msg.pixels[core][y][x] = 0;
					}
				}
			}
			eol = 0;
		}
	}

	e_open(&edev, 0, 0, ROWS, COLS);
	e_write(&emem, 0, 0, 0, &msg, sizeof(msg));
	e_reset_group(&edev);
	e_load_group("epiphany.srec", &edev, 0, 0, ROWS, COLS, E_TRUE);
	nano_wait(0, 100000000);
	clock_gettime(CLOCK_REALTIME, &time);
	time0 = time.tv_sec + time.tv_nsec * 1.0e-9;

	unsigned int frame = 0;
	while (1) {
		for (row = 0; row < ROWS; row++) {
			for (col = 0; col < COLS; col++) {
				core = row * COLS + col;
				while (1) {
					e_read(&emem, 0, 0, (off_t)((char *)&msg.msg_d2h[core] - (char *)&msg), &msg.msg_d2h[core], sizeof(msg_dev2host_t));
					vepiphany[core] = msg.msg_d2h[core].value;
					if (vhost[core] - vepiphany[core] > ((~0u) >> 1)) {
						break;
					}
					nano_wait(0, 1000000);
				}
				vhost[core] = vepiphany[core];
				vcoreid = msg.msg_d2h[core].coreid;
				//printf("%x row:%d col:%d\n", vepiphany[core], (vcoreid >> 6), (vcoreid & 0x3f));
			}
		}
		for (row = 0; row < ROWS; row++) {
			for (col = 0; col < COLS; col++) {
				e_resume(&edev, row, col);
			}
		}
		frame++;
		if (frame > FRAMES) {
			break;
		}
	}

	for (core = 0; core < CORES; core++) {
		while (1) {
			e_read(&emem, 0, 0, (off_t)((char *)&msg.msg_d2h[core] - (char *)&msg), &msg.msg_d2h[core], sizeof(msg_dev2host_t));
			if (msg.msg_d2h[core].finished == 1) {
				break;
			}
			nano_wait(0, 1000000);
		}
		e_read(&emem, 0, 0, (off_t)((char *)&msg.pixels[core] - (char *)&msg), &msg.pixels[core], sizeof(uint32_t[yres][xres]));
	}

	clock_gettime(CLOCK_REALTIME, &time);
	time1 = time.tv_sec + time.tv_nsec * 1.0e-9;
	printf("rows: %d, cols: %d\n", ROWS * yres, COLS * xres);
	printf("frames: %d\n", FRAMES);
	printf("time: %f sec\n", time1 - time0);

	for (row = 0; row < ROWS; row++) {
		for (y = 0; y < yres; y++) {
			for (col = 0; col < COLS; col++) {
				core = row * COLS + col;
				for (x = 0; x < xres; x++) {
					printf("%d ", msg.pixels[core][y][x]);
				}
			}
			printf("\n");
		}
	}

	e_close(&edev);
	e_free(&emem);
	e_finalize();
	return 0;
}
